#include "boader_zone.h"
#include "my_cuda.h"

#include <cuda_runtime.h>
#include <cstdio>
#include <iostream>
#include <mpi.h>
#include <new>

using std::bad_alloc;
using std::cout;
using std::endl;


template <class type> int get_handle(type *buffer, int ptr[], cudaIpcMemHandle_t handle[], int neibpetot);
template <class type> int neibpe_allgather(type src, type peer[], boader_zone *import_table, boader_zone *export_table, MPI_Comm solver_comm);
template <class type> int neibpe_alltoall(type src[], type peer[], boader_zone *import_table, boader_zone *export_table, MPI_Comm solver_comm);
int open_handle(boader_zone *table, int device, const int peerDevice[], const cudaIpcMemHandle_t peerHandle[], MPI_Comm solver_comm);

boader_zone::boader_zone()
{
	boader_zone_init();
}

boader_zone::~boader_zone()
{
	if( boader_zone_delete() ){
		cout << "exit in crs destructor." << endl;
	}
}


int boader_zone::boader_zone_new(int n, int nnz)
{
	this->neibpetot = n;
	this->nnz       = nnz;

	try{
		neibpe = new int[ n ];
		ptr    = new int[n+1];
		index  = new int[nnz];
	}
	catch(bad_alloc)
	{
		cout << "crs bad alloc expection : " << n << " " << nnz << endl;
		return 1;
	}
	catch(...)
	{
		cout << "crs new expection." << endl;
		return 1;
	}
	new_done = true;
	return 0;
}



int boader_zone::boader_zone_init()
{
	neibpetot = 0;
	nnz       = 0;

	neibpe = 0;
	ptr    = 0;
	index  = 0;

	dev_index = 0;

	buffer = 0;
	host_buffer = 0;
	neibpe_buffer = 0;

	status  = 0;
	request = 0;
	comm    = 0;

	new_done = false;

	return 0;
}


int boader_zone::boader_zone_delete()
{
	int i;
	try{
		if(new_done){
			delete[] (neibpe);
			delete[] (ptr   );
			delete[] (index );

			new_done = false;
		}

		if(dev_index) my_cudaFree(dev_index);
		if(buffer)    my_cudaFree(buffer);

		if(host_buffer) delete[] host_buffer;

		if(neibpe_buffer){
			for(i=0; i<neibpetot; i++){
				if(neibpe_buffer[i]) cudaIpcCloseMemHandle(neibpe_buffer[i]);
			}
			delete[] neibpe_buffer;
		}

		if(status)  delete[] status;
		if(request) delete[] request;

//		cudaStreamDestroy(stream);

		boader_zone_init();
	}
	catch(...)
	{
		cout << "crs delete expection." << endl;
		return 1;
	}

	return 0;
}


int set_buffer(boader_zone *outer, boader_zone *inner, MPI_Comm solver_comm)
{
	int device;
	int *outer_peerDevice, *inner_peerDevice;
	cudaIpcMemHandle_t *outer_handle, *inner_handle, *outer_peerHandle, *inner_peerHandle;

	// set MPI_Comm
	inner->comm = solver_comm;
	outer->comm = solver_comm;

	// create stream
	cudaStreamCreate(&inner->stream);
	cudaStreamCreate(&outer->stream);

	// malloc host buffer
	inner->host_buffer = new double[inner->nnz];
	outer->host_buffer = new double[outer->nnz];

	// malloc status
	inner->status  = new MPI_Status[inner->neibpetot];
	outer->status  = new MPI_Status[outer->neibpetot];

	// malloc request
	inner->request = new MPI_Request[inner->neibpetot];
	outer->request = new MPI_Request[outer->neibpetot];
	
	// malloc buffer
	my_cudaMalloc((void**)&inner->buffer, sizeof(double) * inner->nnz, "buffer");
	my_cudaMalloc((void**)&outer->buffer, sizeof(double) * outer->nnz, "buffer");

#ifndef USE_MPI_IPC
	if(!NO_GPU_DIRECT_FLAG) {
	// get handle
	outer_handle = new cudaIpcMemHandle_t [outer->neibpetot];
	inner_handle = new cudaIpcMemHandle_t [inner->neibpetot];
	get_handle(outer->buffer, outer->ptr, outer_handle, outer->neibpetot);
	get_handle(inner->buffer, inner->ptr, inner_handle, inner->neibpetot);

	// all gather device
	cudaGetDevice(&device);
	outer_peerDevice = new int [outer->neibpetot];
	inner_peerDevice = new int [inner->neibpetot];
	neibpe_allgather(device, outer_peerDevice, outer, inner, solver_comm);
	neibpe_allgather(device, inner_peerDevice, inner, outer, solver_comm);

	// all to all handle
	outer_peerHandle = new cudaIpcMemHandle_t [outer->neibpetot];
	inner_peerHandle = new cudaIpcMemHandle_t [inner->neibpetot];
	neibpe_alltoall(inner_handle, outer_peerHandle, outer, inner, solver_comm);
	neibpe_alltoall(outer_handle, inner_peerHandle, inner, outer, solver_comm);

	// open handle
	open_handle(outer, device, outer_peerDevice, outer_peerHandle, solver_comm);
	open_handle(inner, device, inner_peerDevice, inner_peerHandle, solver_comm);

	delete[] outer_handle;
	delete[] outer_peerDevice;
	delete[] outer_peerHandle;

	delete[] inner_handle;
	delete[] inner_peerDevice;
	delete[] inner_peerHandle;
	}
#endif

	return 0;
}


template <class type>
int get_handle(type *buffer, int ptr[], cudaIpcMemHandle_t handle[], int neibpetot)
{
	int i, j, index, length;
	cudaError_t err;  
	for(i=0; i<neibpetot; i++){
		index  = ptr[ i ];
		length = ptr[i+1] - ptr[ i ];
		for(j=0; j<CUDA_IPC_HANDLE_SIZE; j++) handle[i].reserved[j] = 0;

		if(buffer && length){
			err = cudaIpcGetMemHandle(&handle[i], buffer + index);
			if(err != cudaSuccess){
				cout << " cudaIpcGetMemHandle : " << cudaGetErrorString(err) << endl;
				cout << " handle:" << handle[i].reserved << " buffer:" << buffer << " index:" << index << endl;
				return 1;
			}
		}
	}
	return 0;
}


template <class type>
int neibpe_allgather(type src, type peer[], boader_zone *import_table, boader_zone *export_table, MPI_Comm solver_comm)
{
	int i;
	int target;
	MPI_Status *send_status, *recv_status;
	MPI_Request *send_request, *recv_request;

	//---------- recv peer -------------
	recv_request = new MPI_Request[import_table->neibpetot];
	for(i=0; i<import_table->neibpetot; i++){
		target = import_table->neibpe[i];
		MPI_Irecv(peer + i, sizeof(type), MPI_CHAR, target, 0, solver_comm, recv_request + i);
	}

	//---------- send src -------------
	send_request = new MPI_Request[export_table->neibpetot];
	for(i=0; i<export_table->neibpetot; i++){
		target = export_table->neibpe[i];
		MPI_Isend(&src    , sizeof(type), MPI_CHAR, target, 0, solver_comm, send_request + i);
	}

	//---------- recv wait -------------
	recv_status = new MPI_Status[import_table->neibpetot];
	MPI_Waitall(import_table->neibpetot, recv_request, recv_status);
	delete[] recv_request;
	delete[] recv_status;


	//---------- send wait -------------
	send_status = new MPI_Status[export_table->neibpetot];
	MPI_Waitall(export_table->neibpetot, send_request, send_status);
	delete[] send_request;
	delete[] send_status;

	return 0;
}


template <class type>
int neibpe_alltoall(type src[], type peer[], boader_zone *import_table, boader_zone *export_table, MPI_Comm solver_comm)
{
	int i;
	int target;
	MPI_Status *send_status, *recv_status;
	MPI_Request *send_request, *recv_request;

	//---------- recv peer -------------
	recv_request = new MPI_Request[import_table->neibpetot];
	for(i=0; i<import_table->neibpetot; i++){
		target = import_table->neibpe[i];
		MPI_Irecv(peer + i, sizeof(type), MPI_CHAR, target, 0, solver_comm, recv_request + i);
	}

	//---------- send src -------------
	send_request = new MPI_Request[export_table->neibpetot];
	for(i=0; i<export_table->neibpetot; i++){
		target = export_table->neibpe[i];
		MPI_Isend(src  + i , sizeof(type), MPI_CHAR, target, 0, solver_comm, send_request + i);
	}

	//---------- recv wait -------------
	recv_status = new MPI_Status[import_table->neibpetot];
	MPI_Waitall(import_table->neibpetot, recv_request, recv_status);
	delete[] recv_request;
	delete[] recv_status;


	//---------- send wait -------------
	send_status = new MPI_Status[export_table->neibpetot];
	MPI_Waitall(export_table->neibpetot, send_request, send_status);
	delete[] send_request;
	delete[] send_status;

	return 0;
}


int open_handle(boader_zone *table, int device, const int peerDevice[], const cudaIpcMemHandle_t peerHandle[], MPI_Comm solver_comm)
{
	int i, j;
	int canAccessPeer, device_num, rank, node, neibpe_node;
	double **neibpe_buffer;
	cudaError_t err;

	cudaGetDeviceCount(&device_num);
	MPI_Comm_rank(solver_comm, &rank);
	node = rank / device_num; 

	neibpe_buffer = table->neibpe_buffer = new double*[table->neibpetot];
	for(i=0; i<table->neibpetot; i++){
		neibpe_node = table->neibpe[i] / device_num;
		if(neibpe_node != node){
			canAccessPeer = 0;

		}else{
			// peer device -> device : enable peer access
//			cudaSetDevice(peerDevice[i]);
//			cudaDeviceEnablePeerAccess(device, 0);

			// device -> peer device : enable peer access
//			cudaSetDevice(device);
//			cudaDeviceEnablePeerAccess(peerDevice[i], 0);

			cudaDeviceCanAccessPeer(&canAccessPeer, device, peerDevice[i]);
			for(j=0; peerHandle[i].reserved[j]==0 && j<CUDA_IPC_HANDLE_SIZE; j++);
			if(j == CUDA_IPC_HANDLE_SIZE) canAccessPeer *= 0;
		}


		if(canAccessPeer){
//			cout << " " << rank             << "(" << device        << "/" << device_num << ") <===>";
//			cout << " " << table->neibpe[i] << "(" << peerDevice[i]                      << ")" << endl;
			err = cudaIpcOpenMemHandle((void**)&neibpe_buffer[i], peerHandle[i], cudaIpcMemLazyEnablePeerAccess);
			if(err != cudaSuccess){
				cout << " cudaIpcOpenMemHandle : " << cudaGetErrorString(err) << endl;
				return 1;
			}
		}else{
//			cout << " " << rank             << "(" << device        << "/" << device_num << ") <=x=>";
//			cout << " " << table->neibpe[i] << "(" << peerDevice[i]                      << ")" << endl;
			neibpe_buffer[i] = 0;
		}
	}
	return 0;
}


int send_border_zone(boader_zone *zone, int *order)
{
	int i;
	int *buffer;


	try{
		if( my_cudaMalloc((void**)&zone->dev_index, sizeof(int) * zone->nnz, "dev_index") ) throw;
		buffer = new int[zone->nnz];

		for(i=0; i<zone->nnz; i++){
			buffer[i] = order[ zone->index[i] ];
		}

		if( my_cudaMemcpy(zone->dev_index, buffer, sizeof(int) * zone->nnz, cudaMemcpyHostToDevice) ) throw;

		delete[] buffer;
	}
	catch(...){
		cout << "send_border_zone expection." << endl;
		return 1;
	}

	return 0;
}
