#include "cuda_amg.h"
#include "my_cuda.h"
#include "my_mpi_cuda.h"
#include "smoother.h"

#include <cmath>
#include <iostream>

using std::cout;
using std::endl;

void cuda_amg_setup(v_nodes *vs, int level, MPI_Comm solver_comm)
{
	int i;
	int error_check;
	v_node *v, *col, *row;

	error_check = 0;
	try{
		v = &vs->at(level);

		v->smoother_setup( *v );
		error_check += send_border_zone(&v->area_outer, v->order);
		error_check += send_border_zone(&v->area_inner, v->order);
		error_check += set_buffer(&v->area_outer, &v->area_inner, solver_comm);

		if(level != 0){

			row = &vs->at( level );
			col = &vs->at(level-1);
			error_check += convert_crs_to_icrs(v->r, v->dr, row->b_n, row->b_cnt, row->b_ind, col->order);

			row = &vs->at(level-1);
			col = &vs->at( level );
			error_check += convert_crs_to_icrs(v->p, v->dp, row->b_n, row->b_cnt, row->b_ind, col->order);

			error_check += send_border_zone(&v->aggregate_outer, v->order);
			error_check += send_border_zone(&v->aggregate_inner, v->order);
			error_check += set_buffer(&v->aggregate_outer, &v->aggregate_inner, solver_comm);
		}

		v->n   = v->a.n;
		v->np  = v->a.np;
		v->dn  = 0;
		for(i=0; i<v->da.size(); i++){

			if(i != v->da.size()-1){
				v->dn  += v->da.at(i).np;
			}else{
				v->dn  += v->da.at(i).n ;
			}
		}
		v->dnp = v->dn + (v->np - v->n);

		error_check += my_cudaMalloc((void **)&v->dx, sizeof(double) * v->dnp, "dx malloc");	
		error_check += my_cudaMalloc((void **)&v->db, sizeof(double) * v->dnp, "db malloc");
		error_check += my_cudaMalloc((void **)&v->de, sizeof(double) * v->dnp, "de malloc");

		if(level != 0){
			v->x = new double[v->np];
			v->b = new double[v->np];
		}

		if(v->scl_vec){
		  error_check += my_cudaMalloc((void **)&v->d_scl_vec, sizeof(double) * v->dnp, "d_scl_vec malloc");
		  error_check += my_cuda_host_to_device_array(v->d_scl_vec, v->scl_vec, v->dnp, v->np, v->order);
		}
		
		
		  
		if(error_check){
			throw;
		}
	}
	catch(...){
		cout << "cuda_amg_setup expection." << endl;
	}
}


void cuda_amg_cycle_setup(v_nodes *vs, double *x, double *b)
{
	int i;
	int error_check;
	v_node *v;

	error_check = 0;
	try{

		for(i=0; i<vs->size(); i++){	
			v = &vs->at(i);

			error_check += my_cudaMemset(v->dx, 0, sizeof(double) * v->dnp);
			error_check += my_cudaMemset(v->db, 0, sizeof(double) * v->dnp);
			error_check += my_cudaMemset(v->de, 0, sizeof(double) * v->dnp);
		}
	
		v = &vs->at(0);
		v->x = x;
		v->b = b;
		error_check += my_cuda_host_to_device_array(v->dx, x, v->dnp, v->np, v->order);
		error_check += my_cuda_host_to_device_array(v->db, b, v->dnp, v->np, v->order);
		
		if(error_check){
			throw;
		}
	}
	catch(...){
		cout << "cuda_amg_cycle_setup expection." << endl;
	}
}


void cuda_amg_cycle(v_nodes *vs, double *dx, double *db, int *iter, bool (*ratio_test)(v_nodes&))
{
	int i;
	int level_num, loop;
	double *dx_temp, *db_temp;
	v_node *v, *rear, *front;

	v = &vs->at(0);
	dx_temp = v->dx;
	db_temp = v->db;
	v->dx = dx;
	v->db = db;

	level_num = (int)vs->size();

	loop = 0;
	comm_margin<COMM_MARGIN_EQL>(v->dx, &v->area_outer, &v->area_inner);
	while(ratio_test(*vs) == false && loop < *iter){
		loop++;

		for(i=0; i<level_num-1; i++){
			rear  = &vs->at( i );
			front = &vs->at(i+1);

			if(rear->scl_vec){
			         my_cudaScale   (rear->dx, rear->d_scl_vec, rear->dnp);
			         my_cudaInvScale(rear->db, rear->d_scl_vec, rear->dnp);
			}

			rear->smoother(rear->iter, rear->omega);

			residual_error(rear->da, rear->dx, rear->db, rear->de);
			
			if(rear->scl_vec){
			         my_cudaInvScale   (rear->dx, rear->d_scl_vec, rear->dnp);
			         my_cudaScale      (rear->de, rear->d_scl_vec, rear->dn);
			}


			restriction(front->dr, rear->de, front->db);
			
			comm_margin<COMM_MARGIN_ADD>(front->db, &front->aggregate_inner, &front->aggregate_outer);

			my_cudaMemset(front->dx, 0, sizeof(double) * front->dnp);
		}

		v = &vs->at( level_num-1 );

		if(v->scl_vec){
		        my_cudaScale   (v->dx, v->d_scl_vec, v->dnp);
		        my_cudaInvScale(v->db, v->d_scl_vec, v->dnp);
		}

		v->smoother(v->iter, v->omega);

		if(v->scl_vec){
		        my_cudaInvScale(v->dx, v->d_scl_vec, v->dnp);
		}


		for(i=level_num-1; i>0; i--){
			rear  = &vs->at( i );
			front = &vs->at(i-1);			
			
			comm_margin<COMM_MARGIN_EQL>(rear->dx, &rear->aggregate_outer, &rear->aggregate_inner);

			prolongation(rear->dp, rear->dx, front->dx);

			comm_margin<COMM_MARGIN_EQL>(front->dx, &front->area_outer, &front->area_inner);
			
			if(front->scl_vec){
			         my_cudaScale   (front->dx, front->d_scl_vec, front->dnp);
			}

			front->smoother(front->iter, front->omega);
			
			if(front->scl_vec){
			         my_cudaInvScale   (front->dx, front->d_scl_vec, front->dnp);
			}

			

		}
	};

	*iter = loop;
	
	v = &vs->at(0);
	v->dx = dx_temp;
	v->db = db_temp;
}


void cuda_amg_cycle(v_nodes *vs, double *dx, double *db)
{
	int i;
	int level_num;
	double *dx_temp, *db_temp;
	v_node *v, *rear, *front;
	
	v = &vs->at(0);
	dx_temp = v->dx;
	db_temp = v->db;
	v->dx = dx;
	v->db = db;

	level_num = (int)vs->size();
	
	for(i=0; i<level_num-1; i++){
		rear  = &vs->at( i );
		front = &vs->at(i+1);

		if(rear->scl_vec){
		         my_cudaScale   (rear->dx, rear->d_scl_vec, rear->dnp);
		         my_cudaInvScale(rear->db, rear->d_scl_vec, rear->dnp);
		}
		
		rear->smoother(rear->iter, rear->omega);

		residual_error(rear->da, rear->dx, rear->db, rear->de);

		if(rear->scl_vec){
		         my_cudaInvScale   (rear->dx, rear->d_scl_vec, rear->dnp);
		         my_cudaScale      (rear->de, rear->d_scl_vec, rear->dn);
		}

		restriction(front->dr, rear->de, front->db);
			
		comm_margin<COMM_MARGIN_ADD>(front->db, &front->aggregate_inner, &front->aggregate_outer);

		my_cudaMemset(front->dx, 0, sizeof(double) * front->dnp);
	}

	v = &vs->at( level_num-1 );

	if(v->scl_vec){
	        my_cudaScale   (v->dx, v->d_scl_vec, v->dnp);
	        my_cudaInvScale(v->db, v->d_scl_vec, v->dnp);
	}

	v->smoother(v->iter, v->omega);

	if(v->scl_vec){
	        my_cudaInvScale(v->dx, v->d_scl_vec, v->dnp);
	}


	for(i=level_num-1; i>0; i--){
		rear  = &vs->at( i );
		front = &vs->at(i-1);
		
		comm_margin<COMM_MARGIN_EQL>(rear->dx, &rear->aggregate_outer, &rear->aggregate_inner);

		prolongation(rear->dp, rear->dx, front->dx);

		comm_margin<COMM_MARGIN_EQL>(front->dx, &front->area_outer, &front->area_inner);

		if(front->scl_vec){
		         my_cudaScale   (front->dx, front->d_scl_vec, front->dnp);
		}

		front->smoother(front->iter, front->omega);

		if(front->scl_vec){
		         my_cudaInvScale   (front->dx, front->d_scl_vec, front->dnp);
		}

	}
	
	v = &vs->at(0);
	v->dx = dx_temp;
	v->db = db_temp;
}


void cuda_amg_cycle_finalize(v_nodes *vs, double *x, double *b)
{
	v_node *v;

	v = &vs->at(0);
	v->x = 0;
	v->b = 0;
	my_cuda_device_to_host_array(x, v->dx, v->np, v->dnp, v->order);
}


void cuda_amg_finalize(v_nodes *vs)
{
	size_t i;
	size_t level_num;
	v_node *v;

	level_num = vs->size();
	for(i=0; i<vs->size(); i++){
		v = &vs->at(i);
		v->smoother_finalize( *v );

		if(v->dx) my_cudaFree(v->dx);
		if(v->db) my_cudaFree(v->db);
		if(v->de) my_cudaFree(v->de);

		if(v->d_scl_vec) my_cudaFree(v->d_scl_vec);

		v->dx = 0;
		v->db = 0;
		v->de = 0;
		v->d_scl_vec = 0;

		if(i != 0){
			if(v->x) delete[] v->x;
			if(v->b) delete[] v->b;

			v->x = 0;
			v->b = 0;
		}
		
		v->scl_vec = 0;
		
	}
}
