#include "cuda_amg.h"
#include "cuda_bicgstab.h"
#include "my_cuda.h"
#include "my_mpi_cuda.h"

#include <cmath>
#include <iostream>
#include <map>

using std::cout;
using std::endl;
using std::map;

enum ww{R, RT, P, PT, S, ST, T, V, WW_LENGTH};

map<v_nodes*, double**> dww_map;

void bicgstab_func_0(double *dp, double *dr, double *dv, double beta, double omega, int n);
void bicgstab_func_1(double *ds, double *dr, double *dv, double alpha, int n);
void bicgstab_func_2(double *dx, double *dpt, double *dst, double alpha, double omega, int n);

void cuda_bicgstab_setup(v_nodes *vs, int level, MPI_Comm solver_comm)
{
	int i;
	double **dww;
	v_node *v;

	try{
		cuda_amg_setup(vs, level, solver_comm);

		if(level==0){
			v = &vs->at(0);

			if( dww_map.find(vs) != dww_map.end() ){
				cout << "already call cuda_bicgstab_setup." <<endl;
				throw;

			}else{
				dww = dww_map[vs] = new double*[WW_LENGTH];

				for(i=0; i<WW_LENGTH; i++){
					if( my_cudaMalloc((void **)&dww[i], sizeof(double) * v->dnp, "dww") ){
						throw;
					}
					my_cudaMemset(dww[i], 0, sizeof(double) * v->dnp);
				}

			}
		}
	}
	catch(...){
		cout << "cuda_bicgstab_setup expection." << endl;
	}
}


void cuda_bicgstab_cycle(v_nodes *vs, double *x, double *b, int *iter, bool (*ratio_test)(v_nodes&))
{
	v_node *v;

	int i, n, np, loop;
	double rho, rho1, alpha, beta, omega;
	double c2, cg[2];
	double **dww;

	dww = dww_map[vs];

	cuda_amg_cycle_setup(vs, x, b);
	
	v = &vs->at(0);
	n   = v->dn;
	np  = v->dnp;

	for(i=0; i<WW_LENGTH; i++){
		my_cudaMemset(dww[i], 0, sizeof(double) * np);
	}

	// R = B - AX
	comm_margin<COMM_MARGIN_EQL>(v->dx, &v->area_outer, &v->area_inner);
	residual_error(v->da, v->dx, v->db, dww[R]);

	// RT = R
	my_cudaMemcpy(dww[RT], dww[R], sizeof(double) * n, cudaMemcpyDeviceToDevice);

	loop = 0;
	while(ratio_test(*vs) == false && loop < *iter){
		loop++;

		// RHO = R dot RT
		rho = my_cuda_dot_with_mpi(dww[R], dww[RT], n, v->comm);

		if(loop != 1){
			// BETA = (RHO  / RHO1) * (ALPHA / OMEGA)
			beta = (rho / rho1) * (alpha / omega);

			// P = R + BETA * (P - OMEGA * V)
			bicgstab_func_0(dww[P], dww[R], dww[V], beta, omega, n);
		}else{
			// P = R
			my_cudaMemcpy(dww[P], dww[R], sizeof(double) * n, cudaMemcpyDeviceToDevice);
		}

		// V-Cycle A * PT = P
		comm_margin<COMM_MARGIN_EQL>(dww[P], &v->area_outer, &v->area_inner);
		my_cudaMemset(dww[PT], 0, sizeof(double) * np);
		cuda_amg_cycle(vs, dww[PT], dww[P]);

		// V = A * PT
		comm_margin<COMM_MARGIN_EQL>(dww[PT], &v->area_outer, &v->area_inner);
		mmult(v->da, dww[PT], dww[V]);

		// C2 = RT dot V
		c2 = my_cuda_dot_with_mpi(dww[RT], dww[V], n, v->comm);
		
		// ALPHA = RHO / C2
		alpha = rho / c2;

		// S = R - ALPHA * V
		bicgstab_func_1(dww[S], dww[R], dww[V], alpha, n);

		// V-Cycle A * ST = S
		comm_margin<COMM_MARGIN_EQL>(dww[S], &v->area_outer, &v->area_inner);
		my_cudaMemset(dww[ST], 0, sizeof(double) * np);
		cuda_amg_cycle(vs, dww[ST], dww[S]);

		// T = A * ST
		comm_margin<COMM_MARGIN_EQL>(dww[ST], &v->area_outer, &v->area_inner);
		mmult(v->da, dww[ST], dww[T]);

		// CG[0] = T dot S
		cg[0] = my_cuda_dot_with_mpi(dww[T], dww[S], n, v->comm);

		// CG[1] = T dot T
		cg[1] = my_cuda_dot_with_mpi(dww[T], dww[T], n, v->comm);

		// OMEGA = CG[0] / CG[1]
		omega = cg[0] / cg[1];

		// X = X + ALPHA * PT + OMEGA * ST
		bicgstab_func_2(v->dx, dww[PT], dww[ST], alpha, omega, n);

		// R = S - OMEGA * T
		bicgstab_func_1(dww[R], dww[S], dww[T], omega, n);

		// RHO1 = RHO
		rho1 = rho;

		comm_margin<COMM_MARGIN_EQL>(v->dx, &v->area_outer, &v->area_inner);

	};
	*iter = loop;

	cuda_amg_cycle_finalize(vs, x, b);

	return;
}


void cuda_bicgstab_finalize(v_nodes *vs)
{
	int i;
	double **dww;

	dww = dww_map[vs];

	for(i=0; i<WW_LENGTH; i++){
		my_cudaFree(dww[i]);
	}
	dww_map.erase(vs);

	cuda_amg_finalize(vs);
}
