#include "cuda_idr_jacobi.h"
#include "my_mpi_cuda.h"

#include <cmath>
#include <iostream>
#include <map>

#include <cuda_runtime.h>
#include <device_launch_parameters.h>

#define SP 256

using std::cout;
using std::endl;

__global__
void idr_jacobi_kernel_0(double *s, double *d, double *r, double *dr, double *x, double *dx, double gamma, double omega, int n);
__global__
void idr_jacobi_kernel_1(lu_icrs_part a, double *s, double *r, double *dr, double omega);

__global__
void idr_jacobi_kernel_2(double *t, double *r, double *dr, double gamma, int n);

__global__
void idr_jacobi_kernel_3(lu_icrs_part a, double *dat, double *t);

__global__
void idr_jacobi_kernel_4(double *s, double *d, double *t, double *x, double *dx, double gamma, double omega, int n);


void idr_jacobi(lu_icrs &a, double *x, double *b, double *t, double *dat, double *r, double *p, double *s, double *dr, double *dx, double *gamma, double omega, MPI_Comm comm)
{
	int n;

	n = a[0].n;

	if(false){

		idr_jacobi_kernel_0<<< n / SP + 1, SP >>>(s, a[0].d, r, dr, x, dx, *gamma, omega, n);
	
		idr_jacobi_kernel_1<<< n / SP + 1, SP >>>(a[0], s, r, dr, omega);
	
		*gamma = my_cuda_dot_with_mpi(p, r, n, comm) / my_cuda_dot_with_mpi(p, dr, n, comm);

	}else{
		idr_jacobi_kernel_2<<< n / SP + 1, SP >>>(t, r, dr, *gamma, n);

		idr_jacobi_kernel_3<<< n / SP + 1, SP >>>(a[0], dat, t);

		omega = my_cuda_dot_with_mpi(dat, t, n, comm) / my_cuda_dot_with_mpi(dat, dat, n, comm);
		if(omega > 2.0) omega = 2.0;
		if(omega < 0.1) omega = 0.1;

		idr_jacobi_kernel_4<<< n / SP + 1, SP >>>(s, a[0].d, t, x, dx, *gamma, omega, n);

		idr_jacobi_kernel_1<<< n / SP + 1, SP >>>(a[0], s, r, dr, omega);
	
		*gamma = my_cuda_dot_with_mpi(p, r, n, comm) / my_cuda_dot_with_mpi(p, dr, n, comm);
	}
	return;
}


__global__
void idr_jacobi_kernel_0(double *s, double *d, double *r, double *dr, double *x, double *dx, double gamma, double omega, int n)
{
	const int id  = blockIdx.x * blockDim.x + threadIdx.x;

	if(id < n){
		s[id]  = omega / d[id] * ( r[id] - gamma * dr[id] );
		dx[id] = s[id] - gamma * dx[id];
		x[id]  = x[id] + dx[id];
	}
}


__global__
void idr_jacobi_kernel_1(lu_icrs_part a, double *s, double *r, double *dr, double omega)
{
	int i, index, local_w,  col;
	double temp, local_d, local_s, local_r, val;

	const int id  = blockIdx.x * blockDim.x + threadIdx.x;

	if(id < a.n){
		index = id + a.row;
		local_w = a.cnt[id];
		local_d = a.d[id];
		local_s = s[index];
		local_r = r[index];

		temp = 0.0;

		index = id;
		for(i=0; i<local_w; i++){
			col = a.col[index];
			val = a.val[index];

			temp += val * s[col];

			index += a.np;
		}
		temp += (1.0 - 1.0 / omega) * local_d * local_s;
		
		index = id + a.row;
		dr[index] = temp - local_r;
		r[index]  = temp;
	}
}


__global__
void idr_jacobi_kernel_2(double *t, double *r, double *dr, double gamma, int n)
{
	const int id  = blockIdx.x * blockDim.x + threadIdx.x;

	if(id < n){
		t[id] =  r[id] - gamma * dr[id];
	}
}


__global__
void idr_jacobi_kernel_3(lu_icrs_part a, double *dat, double *t)
{
	int i, index, local_w,  col;
	double temp, local_d, local_t, val;

	const int id  = blockIdx.x * blockDim.x + threadIdx.x;

	if(id < a.n){
		index = id + a.row;
		local_w = a.cnt[id];
		local_d = a.d[id];
		local_t = t[index];

		val = 0.0;

		index = id;
		for(i=0; i<local_w; i++){
			col = a.col[index];
			val = a.val[index];

			temp += val * t[col];

			index += a.np;
		}
		temp += local_d * local_t;
		temp = temp / local_d;
		
		index = id + a.row;
		dat[index] = temp;
	}
}


__global__
void idr_jacobi_kernel_4(double *s, double *d, double *t, double *x, double *dx, double gamma, double omega, int n)
{
	const int id  = blockIdx.x * blockDim.x + threadIdx.x;

	if(id < n){
		s[id]  = omega / d[id] * t[id];
		dx[id] = s[id] - gamma * dx[id];
		x[id]  = x[id] + dx[id];
	}
}

