#include "cuda_amg.h"
#include "cuda_bicgstab.h"
#include "gpu_solver.h"
#include "my_cuda.h"
#include "my_mpi_cuda.h"
#include "smoother.h"
#include "output_log.h"

#include <cmath>
#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <iostream>
#include <map>
#include <vector>

#include <mpi.h>

using std::cout;
using std::endl;
using std::map;

class gpu_solver_status
{
public:
	v_nodes *vs;
	int rank;
	int solver_select;
	bool use_direct_solver;

	bool (*ratio_test)(v_nodes& vs);
	double *terminal_vec;

	double terminal_val;
	double first_residual;
	bool first;

	gpu_solver_status()
	{
		terminal_vec    = 0;
		terminal_val    = 0;
		first_residual  = 0;
		first           = true;
	}
};
bool ratio_test_vec(v_nodes& vs);
bool ratio_test_val(v_nodes& vs);

namespace 
{
	map<MPI_Comm, gpu_solver_status> status_map;
}


void notify_level_(int *level_num, int *rank)
{
	int i;
	MPI_Comm solver_comm;
	gpu_solver_status *status;

	solver_comm = MPI_COMM_WORLD;

	if( my_cudaSetDevice(*rank) ){
		MPI_Abort(solver_comm, 1);
		exit(1);
	}

	if(status_map.find(solver_comm) != status_map.end()){
		cout << "already call notify_level_ with MPI_Comm " << solver_comm << "." <<endl;
		return;
	}

	status_map[solver_comm] = gpu_solver_status();

	status = &status_map[solver_comm];

	status->rank = *rank;

	status->vs = new v_nodes;

	status->vs->reserve(*level_num);

	for(i=0; i<*level_num; i++){
		status->vs->push_back( v_node() );
		status->vs->at(i).level = i;
		status->vs->at(i).comm  = solver_comm;
	}

	return;
}


void parameter_set_on_gpu_(int *select, double *omega, int* iter_down, int* iter_bottom, int* gs_at_lev1)
{
	int type;
	MPI_Comm solver_comm;
	gpu_solver_status *status;

	solver_comm = MPI_COMM_WORLD;

	status = &status_map[solver_comm];
	
	status->solver_select = *select;

	// set smoother.
	type = *gs_at_lev1;
	smoother_set_on_gpu_(&type);

	// set number of iter
	iter_set_on_gpu_(iter_down, iter_bottom);

	// set omega.
	omega_set_on_gpu_(omega);

	return;
}


void smoother_set_on_gpu_(int *smoother_set)
{
	int i;
	int level_num, set, type;
	v_node *v;
	MPI_Comm solver_comm;
	gpu_solver_status *status;

	solver_comm = MPI_COMM_WORLD;

	status = &status_map[solver_comm];

	level_num = (int)status->vs->size();

	if(*smoother_set < SMOOTHER_SET_TYPE_LENGTH / 2){
		status->use_direct_solver = false;
		set = *smoother_set;
	}else{
		status->use_direct_solver = true;
		set = *smoother_set - SMOOTHER_SET_TYPE_LENGTH / 2;
	}

	for(i=0; i<level_num; i++){
		v = &status->vs->at(i);
		
		// set smoother.
		switch(set){
		case ALL_JC:
			if(i==0){
//				if(status->rank == 0) cout << " select ALL_JC " << endl;
				type = JC;
			}else if(i != level_num-1){
				type = JC;
			}else{
				type = JC;
			}
			break;
		case MCGS_AND_JC:
			if(i==0){
//				if(status->rank == 0) cout << " select MCGS_AND_JC " << endl;
				type = MCGS;
			}else if(i != level_num-1){
				type = JC;
			}else{
				type = JC;
			}
			break;
		case ALL_MCGS:
			if(i==0){
//				if(status->rank == 0) cout << " select ALL_MCGS " << endl;
				type = MCGS;
			}else if(i != level_num-1){
				type = MCGS;
			}else{
				type = MCGS;
			}
			break;
		case IDR_AND_JC:
			if(i==0){
//				if(status->rank == 0) cout << " select IDR_AND_JC " << endl;
				type = IDR;
			}else if(i != level_num-1){
				type = JC;
			}else{
				type = JC;
			}
			break;
		default:
			cout << "please select smoother set !" << endl;
			MPI_Abort(solver_comm, 1);
			exit(1);
		}

		if(i == level_num-1 && status->use_direct_solver){
			int local_nnz, nnz;

			local_nnz = v->area_inner.nnz + v->area_outer.nnz;
			MPI_Allreduce(&local_nnz, &nnz, 1, MPI_INT, MPI_SUM, solver_comm);

			if(nnz == 0){
//				cout << " use LU decomposition " << endl;
				type = LU;
				status->use_direct_solver = true;
			}else{
				status->use_direct_solver = false;
			}
		}

		switch(type){
		case MCGS:
			v->smoother_setup       = multi_color_gauss_seidel_setup;
			v->smoother_kernel      = multi_color_gauss_seidel;
			break;
		case JC:
			v->smoother_setup       = jacobi_setup;
			v->smoother_kernel      = jacobi;
			v->smoother_finalize    = jacobi_finalize;
			break;
		case IDR:
			v->smoother_setup       = idr_jacobi_setup;
			v->smoother_kernel_init = idr_jacobi_init;
			v->smoother_kernel      = idr_jacobi;
			v->smoother_finalize    = idr_jacobi_finalize;
			break;
		case LU:
			v->smoother_setup       = lu_decomposition_setup;
			v->smoother_kernel      = lu_decomposition;
			v->smoother_finalize    = lu_decomposition_finalize;
			break;
		default:
			cout << "please select smoother !" << endl;
			MPI_Abort(solver_comm, 1);
			exit(1);
		}
	}
}


void omega_set_on_gpu_(double *omega)
{
	int i;
	int level_num;
	v_node *v;
	MPI_Comm solver_comm;
	gpu_solver_status *status;

	solver_comm = MPI_COMM_WORLD;

	status = &status_map[solver_comm];

	level_num = (int)status->vs->size();

	for(i=0; i<level_num; i++){
		v = &status->vs->at(i);
		v->omega = *omega;
	}
}


void iter_type_set_on_gpu_(int *iter_type)
{
	int i;
	int level_num;
	v_node *v;
	MPI_Comm solver_comm;
	gpu_solver_status *status;

	solver_comm = MPI_COMM_WORLD;

	status = &status_map[solver_comm];

	level_num = (int)status->vs->size();

	switch(*iter_type){
	case FLAT:
		for(i=0; i<level_num; i++){
			v = &status->vs->at(i);
			if(i != level_num-1){
				v->iter = 2;
			}else{
				v->iter = 30;
			}
		}
		break;
	case BOTTOM_UP:
		for(i=0; i<level_num; i++){
			v = &status->vs->at(i);
			if(i != level_num-1){
				v->iter = i + 1;
			}else{
				v->iter = 30;
			}
		}
		break;
	case BOTTOM_DOWN:
		for(i=0; i<level_num; i++){
			v = &status->vs->at(i);
			if(i != level_num-1){
				v->iter = level_num - i - 1;
			}else{
				v->iter = 30;
			}
		}
		break;
	case NON_ITER:
		for(i=0; i<level_num; i++){
			v = &status->vs->at(i);
			v->iter = 0;
		}
		break;
	default:
		cout << "please select iter_type !" << endl;
		break;
	}
}


void iter_set_on_gpu_(int* iter_down, int* iter_bottom)
{
	int i;
	int level_num;
	v_node *v;
	MPI_Comm solver_comm;
	gpu_solver_status *status;

	solver_comm = MPI_COMM_WORLD;

	status = &status_map[solver_comm];

	level_num = (int)status->vs->size();

	for(i=0; i<level_num; i++){
		v = &status->vs->at(i);
		if( i == 0){
			v->iter = *iter_down;
		}else if(i != level_num-1){
			v->iter = *iter_down;
		}else{
			v->iter = *iter_bottom;
		}
	}
}


void store_mat_on_gpu_(int *Level, 
					  int *N, int *NP, int *NPL, int *NPU, 
					  int *INL, int *INU, 
					  int *IAL, double *AL, 
					  int *IAU, double *AU, 
					  double *D,

					  int *IN_T_N, int *IN_T_NPLU,
					  int *IN_T_NO, int *IN_T_DSP, int *IN_T_COL,
					  int *EX_T_N, int *EX_T_NPLU,
					  int *EX_T_NO, int *EX_T_DSP, int *EX_T_COL,

					  int *R_N, int *P_N, int *R_NP, int *P_NP, int *R_NPLU, int *P_NPLU, 
					  int *R_IN, int *R_CN, double *R_V,
					  int *P_IN, int *P_CN, double *P_V,

					  int *INTER_IN_T_N, int *INTER_IN_T_NPLU,
					  int *INTER_IN_T_NO, int *INTER_IN_T_DSP, int *INTER_IN_T_COL,
					  int *INTER_EX_T_N, int *INTER_EX_T_NPLU,
		                          int *INTER_EX_T_NO, int *INTER_EX_T_DSP, int *INTER_EX_T_COL,
		                          
		                          int *SCL_FL, double *SCL_VEC
)
{
	int i;
	v_node *v;
	lu_crs *a;
	crs *r, *p;
	boader_zone *area_inner, *area_outer, *aggregate_inner, *aggregate_outer;
	MPI_Comm solver_comm;
	gpu_solver_status *status;

	solver_comm = MPI_COMM_WORLD;

	status = &status_map[solver_comm];

	v = &status->vs->at(*Level-1);
	a = &v->a;
	r = &v->r;
	p = &v->p;
	area_outer = &v->area_outer;
	area_inner = &v->area_inner;
	aggregate_outer = &v->aggregate_outer;
	aggregate_inner = &v->aggregate_inner;

	a->n	= *N;
	a->np	= *NP;
	a->nnzl	= *NPL;
	a->nnzu	= *NPU;
	a->inl	= INL;
	a->inu	= INU;
	a->ial	= IAL;
	a->iau	= IAU;
	a->al	= AL;
	a->au	= AU;
	a->d	= D;

	area_outer->neibpetot = *IN_T_N;
	area_outer->nnz    = *IN_T_NPLU;
	area_outer->boader_zone_new(area_outer->neibpetot, area_outer->nnz);
	area_outer->ptr[0] = 0;
	if(area_outer->neibpetot){
		memcpy(area_outer->neibpe, IN_T_NO , sizeof(int) * area_outer->neibpetot);
		memcpy(area_outer->ptr+1 , IN_T_DSP, sizeof(int) * area_outer->neibpetot);
	}
	if(area_outer->nnz){
		memcpy(area_outer->index , IN_T_COL, sizeof(int) * area_outer->nnz);
	}

	area_inner->neibpetot = *EX_T_N;
	area_inner->nnz    = *EX_T_NPLU;
	area_inner->boader_zone_new(area_inner->neibpetot, area_inner->nnz);
	area_inner->ptr[0] = 0;
	if(area_inner->neibpetot){
		memcpy(area_inner->neibpe, EX_T_NO , sizeof(int) * area_inner->neibpetot);
		memcpy(area_inner->ptr+1 , EX_T_DSP, sizeof(int) * area_inner->neibpetot);
	}
	if(area_inner->nnz){
		memcpy(area_inner->index , EX_T_COL, sizeof(int) * area_inner->nnz);
	}

	if(*Level-1 != 0){
		r->n    = *R_NP;
		r->nnz  = *R_NPLU;
		r->ptr  = R_IN;
		r->col  = R_CN;
		r->val  = R_V;

		p->n    = *P_N;
		p->nnz  = *P_NPLU;
		p->ptr  = P_IN;
		p->col  = P_CN;
		p->val  = P_V;

		aggregate_outer->neibpetot = *INTER_IN_T_N;
		aggregate_outer->nnz    = *INTER_IN_T_NPLU;
		aggregate_outer->boader_zone_new(aggregate_outer->neibpetot, aggregate_outer->nnz);
		aggregate_outer->ptr[0] = 0;
		if(aggregate_outer->neibpetot){
			memcpy(aggregate_outer->neibpe, INTER_IN_T_NO , sizeof(int) * aggregate_outer->neibpetot);
			memcpy(aggregate_outer->ptr+1 , INTER_IN_T_DSP, sizeof(int) * aggregate_outer->neibpetot);
		}
		if(aggregate_outer->nnz){
			memcpy(aggregate_outer->index , INTER_IN_T_COL, sizeof(int) * aggregate_outer->nnz);
		}
	
		aggregate_inner->neibpetot = *INTER_EX_T_N;
		aggregate_inner->nnz    = *INTER_EX_T_NPLU;
		aggregate_inner->boader_zone_new(aggregate_inner->neibpetot, aggregate_inner->nnz);
		aggregate_inner->ptr[0] = 0;
		if(aggregate_inner->neibpetot){
			memcpy(aggregate_inner->neibpe, INTER_EX_T_NO , sizeof(int) * aggregate_inner->neibpetot);
			memcpy(aggregate_inner->ptr+1 , INTER_EX_T_DSP, sizeof(int) * aggregate_inner->neibpetot);
		}
		if(aggregate_inner->nnz){
			memcpy(aggregate_inner->index , INTER_EX_T_COL, sizeof(int) * aggregate_inner->nnz);
		}
	}

	// modified col--
	for(i=0; i<a->nnzl; i++) a->ial[i]--;
	for(i=0; i<a->nnzu; i++) a->iau[i]--;
	for(i=0; i<area_outer->nnz; i++) area_outer->index[i]--;
	for(i=0; i<area_inner->nnz; i++) area_inner->index[i]--;

	if(*Level-1 != 0){
		for(i=0; i<r->nnz; i++) r->col[i]--;
		for(i=0; i<p->nnz; i++) p->col[i]--;
		for(i=0; i<aggregate_outer->nnz; i++) aggregate_outer->index[i]--;
		for(i=0; i<aggregate_inner->nnz; i++) aggregate_inner->index[i]--;
	}

	if(*SCL_FL) v->scl_vec = SCL_VEC;

	switch(status->solver_select){
	case AMG:
		cuda_amg_setup(status->vs, *Level-1, solver_comm);
		break;
	case AMG_BiCGSTAB:
		cuda_bicgstab_setup(status->vs, *Level-1, solver_comm);
		break;
	default:
		cout <<  "please select AMG or BiCGSTAB !" << endl;
		break;
	}

	// modified col++
	for(i=0; i<a->nnzl; i++) a->ial[i]++;
	for(i=0; i<a->nnzu; i++) a->iau[i]++;
	for(i=0; i<area_outer->nnz; i++) area_outer->index[i]++;
	for(i=0; i<area_inner->nnz; i++) area_inner->index[i]++;
	if(*Level-1 != 0){
		for(i=0; i<r->nnz; i++) r->col[i]++;
		for(i=0; i<p->nnz; i++) p->col[i]++;
		for(i=0; i<aggregate_outer->nnz; i++) aggregate_outer->index[i]++;
		for(i=0; i<aggregate_inner->nnz; i++) aggregate_inner->index[i]++;
	}

	
	
	return;
}


void amgs_terminal_vec_set_on_gpu_(double *res, int *row)
{
	v_node *v;
	MPI_Comm solver_comm;
	gpu_solver_status *status;

	solver_comm = MPI_COMM_WORLD;

	status = &status_map[solver_comm];

	v = &status->vs->at(0);
	if(status->terminal_vec) my_cudaFree(status->terminal_vec);
	my_cudaMalloc( (void**)&status->terminal_vec, sizeof(double) * v->dn, "terminal_vec");
	my_cudaMemset( status->terminal_vec, 0, sizeof(double) * v->dn);
	my_cuda_host_to_device_array(status->terminal_vec, res, v->dn, *row, v->order);

	status->ratio_test = ratio_test_vec;

	return;
}


void amgs_terminal_val_set_on_gpu_(double *res)
{
	MPI_Comm solver_comm;
	gpu_solver_status *status;

	solver_comm = MPI_COMM_WORLD;

	status = &status_map[solver_comm];

	status->ratio_test = ratio_test_val;

	status->terminal_val = *res;
	return;
}

void solve_on_gpu_(double *x, double *b, int *iter)
{
	int temp, coarsest_lv;
	double *dx, *db;
	v_nodes *vs;
	MPI_Comm solver_comm;
	gpu_solver_status *status;

	solver_comm = MPI_COMM_WORLD;

	status = &status_map[solver_comm];

#ifdef OUTPUT_LOG
	comm_log_clear();
	residual_log_clear();
#endif

	vs = status->vs;

	if(status->use_direct_solver){
		coarsest_lv = vs->size() - 1;
		temp = vs->at( coarsest_lv ).iter;
		vs->at( coarsest_lv ).iter = 1;
	}

	switch(status->solver_select){
	case AMG:
		cuda_amg_cycle_setup(vs, x, b);

		dx = vs->at(0).dx;
		db = vs->at(0).db;
		cuda_amg_cycle(vs, dx, db, iter, status->ratio_test);

		cuda_amg_cycle_finalize(vs, x, b);
		break;
	case AMG_BiCGSTAB:
		cuda_bicgstab_cycle(vs, x, b, iter, status->ratio_test);
		break;
	default:
		cout << "unknown error." << endl;
		break;
	}

	if(status->use_direct_solver){
		coarsest_lv = vs->size() - 1;
		vs->at( coarsest_lv ).iter = temp;
	}
	
	return;
}


void clear_mat_on_gpu_(void)
{
	MPI_Comm solver_comm;
	gpu_solver_status *status;

	solver_comm = MPI_COMM_WORLD;

	status = &status_map[solver_comm];

	switch(status->solver_select){
	case AMG:
		cuda_amg_finalize(status->vs);
		break;
	case AMG_BiCGSTAB:
		cuda_bicgstab_finalize(status->vs);
		break;
	default:
		cout << "unknown error." << endl;
		break;
	}

	status->vs->clear();
	delete status->vs;

	if(status->terminal_vec){
		my_cudaFree(status->terminal_vec);
		status->terminal_vec=0;
	}

	my_cudaGetLastError();
	fflush(stdout);
	
	// added
	status_map.erase(solver_comm);
	
	return;
}


void *get_amgs_v_nodes()
{
	MPI_Comm solver_comm;
	gpu_solver_status *status;

	solver_comm = MPI_COMM_WORLD;

	status = &status_map[solver_comm];

	return (void*)status->vs;
}


bool ratio_test_vec(v_nodes& vs)
{
	bool flag;
	gpu_solver_status *status;

	status = &status_map[vs[0].comm];

	residual_error(vs[0].da, vs[0].dx, vs[0].db, vs[0].de);
	flag = my_cuda_vector_less_with_mpi(vs[0].de, status->terminal_vec, vs[0].dn, vs[0].comm);

	return flag;
}


bool ratio_test_val(v_nodes& vs)
{
	double residual;
	gpu_solver_status *status;
	status = &status_map[vs[0].comm];

	if(status->first){
		status->first_residual = get_two_norm_residual_with_mpi(vs[0].da, vs[0].dx, vs[0].db, vs[0].comm);
		status->first = false;
	}

	residual = get_two_norm_residual_with_mpi(vs[0].da, vs[0].dx, vs[0].db, vs[0].comm) / status->first_residual;

#ifdef OUTPUT_LOG
	residual_log_input(residual);
#endif
#ifdef PRINT_REZ
	if(status->rank == 0) cout << std::scientific <<  " " << residual << std::fixed << endl;
#endif

	if(residual < status->terminal_val || residual > 1.0E+5){
		status->first_residual = 0;
		status->first = true;
		return true;
	}else{
		return false;
	}
}
