#include "lu_icrs.h"
#include "my_cuda.h"

#include <cuda_runtime.h>
#include <device_launch_parameters.h>

#define SP 256
#define SM 16

template <unsigned int blockSize, class type>
__device__ void warp_reduce(volatile type sdata[], int tid);
template <unsigned int blockSize, class type>
__device__ void shared_reduction(type sdata[], int tid);

__global__ void get_two_norm_residual_kernel(lu_icrs_part a, double *x, double *b, double *norm);
__global__ void mmult_kernel(lu_icrs_part a, double *b, double *c);
__global__ void residual_error_kernel(lu_icrs_part a, double *x, double *b, double *r);


double get_two_norm_residual(lu_icrs &a, double *x, double *b)
{
	int i, size;
	double norm;
	double *dsum, *sum;

	size = a.size();

	sum = new double[SM];
	my_cudaMalloc((void**)&dsum, sizeof(double) * SM, "2 norm malloc");

	for(i=0; i<size; i++){
		get_two_norm_residual_kernel<<< SM, SP>>>(a[i], x, b, dsum);
	}
	my_cudaMemcpy(sum, dsum, sizeof(double) * SM, cudaMemcpyDeviceToHost);

	norm = 0.0;
	for(i=0; i<SM; i++){
		norm += sum[i];
	}

	my_cudaFree(dsum);
	delete[] sum;

	return sqrt(norm);
}


void mmult(lu_icrs &a, double *b, double *c)
{
	int i, size;
	size = a.size();
	for(i=0; i<size; i++){
		mmult_kernel<<< a[i].n / SP + 1, SP >>>(a[i], b, c);
	}
}


void residual_error(lu_icrs &a, double *x, double *b, double *e)
{
	int i, size;
	size = a.size();
	for(i=0; i<size; i++){
		residual_error_kernel<<< a[i].n / SP + 1, SP >>>(a[i], x, b, e);
	}
}


template <unsigned int blockSize, class type>
__device__
void warp_reduce(volatile type sdata[], int tid){
	if(blockSize >= 64) sdata[tid] += sdata[tid + 32];
	if(blockSize >= 32) sdata[tid] += sdata[tid + 16];
	if(blockSize >= 16) sdata[tid] += sdata[tid +  8];
	if(blockSize >=  8) sdata[tid] += sdata[tid +  4];
	if(blockSize >=  4) sdata[tid] += sdata[tid +  2];
	if(blockSize >=  2) sdata[tid] += sdata[tid +  1];
}


template <unsigned int blockSize, class type>
__device__
void shared_reduction(type sdata[], int tid)
{

	if(blockSize >= 512){ if(tid <256){ sdata[tid] += sdata[tid + 256];} __syncthreads();}
	if(blockSize >= 256){ if(tid <128){ sdata[tid] += sdata[tid + 128];} __syncthreads();}
	if(blockSize >= 128){ if(tid < 64){ sdata[tid] += sdata[tid +  64];} __syncthreads();}

	if(tid < 32) warp_reduce<blockSize>(sdata, tid);
}


__global__
void get_two_norm_residual_kernel(lu_icrs_part a, double *x, double *b, double *norm)
{
	__shared__ double s_norm[SP];
	__shared__ double s_val[SP];
	__shared__ double s_x[SP];
	__shared__ int    s_col[SP];

	int i, local_row, index, local_w;
	double temp, local_d, local_b, local_x;

	const int tid = threadIdx.x;
	const int id  = blockIdx.x * blockDim.x + threadIdx.x;

	s_norm[tid] = 0.0;

	local_row = id;
	while(local_row < a.n){
		index = local_row + a.row;

		local_w = a.cnt[local_row];
		local_d = a.d[local_row];
		local_x = x[index];
		local_b = b[index];

		temp = 0.0;

		index = local_row;
		for(i=0; i<local_w; i++){
			s_col[tid] = a.col[index];
			s_val[tid] = a.val[index];

			index += a.np;

			s_x[tid] = x[s_col[tid]];

			temp -= s_val[tid] * s_x[tid];
		}
		temp -= local_d * local_x;
		temp += local_b;

		s_norm[tid] += temp * temp;

		local_row += SM*SP;
	}
	__syncthreads();

	shared_reduction<SP>(s_norm, tid);

	if(!tid) norm[blockIdx.x] += s_norm[0];
}

__global__
void mmult_kernel(lu_icrs_part a, double *b, double *c)
{
	__shared__ double s_val[SP];
	__shared__ double s_b[SP];
	__shared__ int    s_col[SP];

	int i, index, local_w;
	double temp, local_d, local_b;

	const int tid = threadIdx.x;
	const int id  = blockIdx.x * blockDim.x + threadIdx.x;

	if(id < a.n){
		index = id + a.row;
		local_w = a.cnt[id];
		local_d = a.d[id];
		local_b = b[index];

		temp = 0.0;

		index = id;
		for(i=0; i<local_w; i++){
			s_col[tid] = a.col[index];
			s_val[tid] = a.val[index];

			index += a.np;

			s_b[tid] = b[s_col[tid]];

			temp += s_val[tid] * s_b[tid];
		}
		temp += local_d * local_b;

		index = id + a.row;
		c[index] = temp;
	}
}


__global__
void residual_error_kernel(lu_icrs_part a, double *x, double *b, double *r)
{
	__shared__ double s_val[SP];
	__shared__ double s_x[SP];
	__shared__ int    s_col[SP];

	int i, index, local_w;
	double temp, local_d, local_b, local_x;

	const int tid = threadIdx.x;
	const int id  = blockIdx.x * blockDim.x + threadIdx.x;

	if(id < a.n){
		index = id + a.row;
		local_w = a.cnt[id];
		local_d = a.d[id];
		local_x = x[index];
		local_b = b[index];

		temp = 0.0;

		index = id;
		for(i=0; i<local_w; i++){
			s_col[tid] = a.col[index];
			s_val[tid] = a.val[index];

			index += a.np;

			s_x[tid] = x[s_col[tid]];

			temp -= s_val[tid] * s_x[tid];
		}
		temp -= local_d * local_x;

		index = id + a.row;
		r[index] = temp + local_b;
	}
}
