#include "my_cuda.h"

#include <cuda_runtime.h>
#include <device_launch_parameters.h>
#include <iostream>
#include <list>
#include <stdio.h>
#include <stdlib.h>

#define SP 256
#define SM 16

using std::cout;
using std::endl;



template <unsigned int blockSize, class type> __device__ void warp_reduce(volatile type sdata[], int tid);
template <unsigned int blockSize>             __device__ void warp_reduce(volatile bool sdata[], int tid);
template <unsigned int blockSize, class type> __device__ void shared_reduction(type sdata[], int tid);
template <unsigned int blockSize>             __device__ void shared_reduction(bool sdata[], int tid);

class malloc_item
{
public:
	void *pointer;
	size_t size;
	char str[1024];

	malloc_item(void* pointer, const size_t size, const char str[])
	{
		int i;
		this->pointer = pointer;
		this->size    = size;

		for(i=0; str[i]; i++){
			this->str[i] = str[i];
		}
		this->str[i] = '\0';
	}
};

int cuda_malloc_count = 0;
std::list<malloc_item> cuda_malloc_list;

int my_cudaMalloc(void **devptr, size_t count, char str[])
{
	cudaError_t error_text;

	error_text = cudaMalloc(devptr, count);

	if (error_text != cudaSuccess) {
		cout << "cudaMalloc : " << cudaGetErrorString(error_text) << endl;	
		cout << "devptr : " << devptr << " ";
		cout << "count  : " << count  << " ";
		cout << "str    : " << str    << " ";
		cout << endl;
		return 1;
	}
	
	cudaMemset(*devptr, 0, count);

	if(count){
		cuda_malloc_count++;
		cuda_malloc_list.push_back( malloc_item(*devptr, count, str) );
	}

	return 0;
}


int my_cudaFree(void *devptr)
{
	cudaError_t error_text;

	error_text = cudaFree(devptr);

	if (error_text != cudaSuccess) {
		cout << "cudaFree : " << cudaGetErrorString(error_text) << endl;
		cout << "devptr : " << devptr << endl;
		return 1;
	}

	std::list<malloc_item>::iterator it;
	for(it=cuda_malloc_list.begin(); it!=cuda_malloc_list.end();){
		if((*it).pointer == devptr){
			it = cuda_malloc_list.erase(it);
			cuda_malloc_count--;
			continue;
		}
		it++;
	}

	return 0;
}


int my_cudaMemcpy(void *dst, const void *src, size_t count, cudaMemcpyKind kind)
{
	cudaError_t error_text;

	error_text = cudaMemcpy(dst, src, count, kind);

	if (error_text != cudaSuccess) {
		cout << "cudaMemcpy : " << cudaGetErrorString(error_text) << endl;
		cout << "dst   : "   << dst   << " ";
		cout << "src   : "   << src   << " ";
		cout << "count : " << count << " ";
		cout << endl;

		return 1;
	}
	return 0;
}


int my_cudaMemset(void *devptr, int c, size_t count)
{
	cudaError_t error_text;

	error_text = cudaMemset(devptr, c, count);

	if (error_text != cudaSuccess) {
		cout << "cudaMemset : " << cudaGetErrorString(error_text) << endl;
		cout << "devptr : "   << devptr   << " ";
		cout << "c      : "   << c   << " ";
		cout << "count  : " << count << " ";
		cout << endl;

		return 1;
	}
	return 0;
}


int my_cudaSetDevice(int device)
{
	int device_num;

	cudaGetDeviceCount(&device_num);

	if(device_num == 0){
		cout << " ### ERROR : cudaGetDeviceCount return 0. ### " << endl;
		return 1;
	}

	device = device % device_num;

	cudaSetDevice(device);

	return 0;
}


int my_cudaGetLastError(void)
{
	cudaError_t err = cudaGetLastError();

	if (err != cudaSuccess) {
		cout << "cudaGetLastError : " << cudaGetErrorString(err) << endl;
	}


	std::list<malloc_item>::iterator it;

	if(cuda_malloc_count){
		printf("cuda malloc count:%d\n", cuda_malloc_count);
		for(it=cuda_malloc_list.begin(); it!=cuda_malloc_list.end(); it++){
			printf("%p %8u %s\n", (*it).pointer, (*it).size, (*it).str);
		}
	}

	if (err != cudaSuccess) {
		return 1;
	}else{
		return 0;
	}
}


__global__
void my_cuda_dot_kernel(double *A, double *B, int N, double *V)
{
	__shared__ double s_V[SP];

	const int tid = threadIdx.x;
	const int id  = blockIdx.x * blockDim.x + threadIdx.x;

	int index = id;

	/* ɤν */
	s_V[tid] = 0.0;

	/* ʬ */
	while(index < N){
		s_V[tid] += A[index] * B[index];
		index += SM*SP;
	}
	__syncthreads();


	/* ɥ */
	shared_reduction<SP>(s_V, tid);


	/* Х¤ư */
	if(tid == 0)
		V[blockIdx.x] = s_V[0];
}


double my_cuda_dot(double *a, double *b, int n)
{
	int i;
	double sum, dot[SM];
	double *ddot;

	my_cudaMalloc((void**)&ddot, sizeof(double) * SM, "dot malloc");

	my_cuda_dot_kernel<<< SM, SP>>>(a, b, n, ddot);

	my_cudaMemcpy(dot, ddot, sizeof(double) * SM, cudaMemcpyDeviceToHost);

	sum = 0.0;
	for(i=0; i<SM; i++){
		sum += dot[i];
	}

	my_cudaFree(ddot);
	return sum;
}


__global__
void my_cuda_scale_kernel(double *a, double *b, int n)
{
	const int tid = threadIdx.x;
	const int id  = blockIdx.x * blockDim.x + threadIdx.x;
	
	int index = id;
		
	while(index < n){
	        a[index] *= b[index];
		index += SM*SP;
	}
}


int my_cudaScale(double *a, double *scl, int n)
{
        my_cuda_scale_kernel<<< SM, SP >>>(a, scl, n);
	return 0;
}

__global__
void my_cuda_inv_scale_kernel(double *a, double *b, int n)
{
	const int tid = threadIdx.x;
	const int id  = blockIdx.x * blockDim.x + threadIdx.x;
	
	int index = id;
		
	while(index < n){
	        a[index] /= b[index];
		index += SM*SP;
	}
}


int my_cudaInvScale(double *a, double *scl, int n)
{
        my_cuda_inv_scale_kernel<<< SM, SP >>>(a, scl, n);
	return 0;

}

__global__ void vector_less_kernel(double *x, double *y, int n, bool *v)
{
	__shared__ bool sv[SP];

	/* ᥤid,åidγ*/
	const int tid = threadIdx.x;
	const int id = blockIdx.x * blockDim.x + threadIdx.x;
	int index = id;

	/* ɤν */
	sv[tid] = true;

	/* ʬ */
	while(index < n){
		sv[tid] = sv[tid] && (fabs(x[index]) <= y[index]);
		index += SM*SP;
	}
	__syncthreads();


	/* ɥ */
	shared_reduction<SP>(sv, tid);


	/* ХӤư */
	if(tid == 0)
		v[blockIdx.x] = sv[0];
}

bool my_cuda_vector_less(double *dx, double *dy, int n)
{
	int i;
	bool sum, dot[SM];
	bool *ddot;

	my_cudaMalloc((void**)&ddot, sizeof(bool) * SM, "vector less malloc");

	vector_less_kernel<<< SM, SP>>>(dx, dy, n, ddot);

	my_cudaMemcpy(dot, ddot, sizeof(bool) * SM, cudaMemcpyDeviceToHost);

	sum = true;
	for(i=0; i<SM; i++){
		sum = sum && dot[i];
	}

	my_cudaFree(ddot);

	return sum;
}


template <unsigned int blockSize, class type>
__device__
void warp_reduce(volatile type sdata[], int tid){
	if(blockSize >= 64) sdata[tid] += sdata[tid + 32];
	if(blockSize >= 32) sdata[tid] += sdata[tid + 16];
	if(blockSize >= 16) sdata[tid] += sdata[tid +  8];
	if(blockSize >=  8) sdata[tid] += sdata[tid +  4];
	if(blockSize >=  4) sdata[tid] += sdata[tid +  2];
	if(blockSize >=  2) sdata[tid] += sdata[tid +  1];
}


template <unsigned int blockSize, class type>
__device__
void shared_reduction(type sdata[], int tid)
{

	if(blockSize >= 512){ if(tid <256){ sdata[tid] += sdata[tid + 256];} __syncthreads();}
	if(blockSize >= 256){ if(tid <128){ sdata[tid] += sdata[tid + 128];} __syncthreads();}
	if(blockSize >= 128){ if(tid < 64){ sdata[tid] += sdata[tid +  64];} __syncthreads();}

	if(tid < 32) warp_reduce<blockSize>(sdata, tid);
}


template <unsigned int blockSize>
__device__
void warp_reduce(volatile bool sdata[], int tid){
	if(blockSize >= 64) sdata[tid] = sdata[tid] && sdata[tid + 32];
	if(blockSize >= 32) sdata[tid] = sdata[tid] && sdata[tid + 16];
	if(blockSize >= 16) sdata[tid] = sdata[tid] && sdata[tid +  8];
	if(blockSize >=  8) sdata[tid] = sdata[tid] && sdata[tid +  4];
	if(blockSize >=  4) sdata[tid] = sdata[tid] && sdata[tid +  2];
	if(blockSize >=  2) sdata[tid] = sdata[tid] && sdata[tid +  1];
}


template <unsigned int blockSize>
__device__
void shared_reduction(bool sdata[], int tid)
{
	if(blockSize >= 512){ if(tid <256){ sdata[tid] = sdata[tid] && sdata[tid + 256];} __syncthreads();}
	if(blockSize >= 256){ if(tid <128){ sdata[tid] = sdata[tid] && sdata[tid + 128];} __syncthreads();}
	if(blockSize >= 128){ if(tid < 64){ sdata[tid] = sdata[tid] && sdata[tid +  64];} __syncthreads();}

	if(tid < 32) warp_reduce<blockSize>(sdata, tid);
}
