#include "lab_matrix.h"
#include "my_cuda.h"
#include "my_mpi_cuda.h"

#include <cmath>
#include <cuda_runtime.h>
#include <iostream>
#include <mpi.h>


#ifdef OUTPUT_LOG
#include "output_log.h"
#endif


// #define NO_GPU_DIRECT_FLAG 0

using std::cout;
using std::endl;

void put_into_the_buffer(double *buffer, const double *src, const int *index, const int n);
void put_into_the_data(double *data, const double *buffer, const int *index, const int n);
void add_into_the_data(double *data, const double *buffer, const int *index, const int n);

template<comm_margin_opp opp>
void comm_margin(double data[], boader_zone *import_table, boader_zone *export_table)
{
	int i;
	int index, length, target;

	//---------- send margin -------------
	put_into_the_buffer(export_table->buffer, data, export_table->dev_index, export_table->nnz);

	for(i=0; i<export_table->neibpetot; i++){
		index  = export_table->ptr[ i ];
		length = export_table->ptr[i+1] - export_table->ptr[i];
		target = export_table->neibpe[i];

#ifndef USE_MPI_IPC
		if(export_table->neibpe_buffer[i] == 0 || NO_GPU_DIRECT_FLAG){
			my_cudaMemcpy(export_table->host_buffer + index, export_table->buffer + index, sizeof(double) * length, cudaMemcpyDeviceToHost);
			MPI_Isend(export_table->host_buffer + index, length, MPI_DOUBLE, target, 0, export_table->comm, export_table->request + i);
		}else{
			cudaThreadSynchronize();
			MPI_Isend(&export_table->signal, 1, MPI_CHAR, target, 0, export_table->comm, export_table->request + i);
		}
#else
		MPI_Isend(export_table->buffer + index, length, MPI_DOUBLE, target, 0, export_table->comm, export_table->request + i);
#endif
	}

	//---------- recv margin -------------
	for(i=0; i<import_table->neibpetot; i++){
		index  = import_table->ptr[ i ];
		length = import_table->ptr[i+1] - import_table->ptr[i];
		target = import_table->neibpe[i];
	
#ifndef USE_MPI_IPC
		if(import_table->neibpe_buffer[i] == 0 || NO_GPU_DIRECT_FLAG){
			MPI_Irecv(import_table->host_buffer + index, length, MPI_DOUBLE, target, 0, import_table->comm, import_table->request + i);
		}else{
			MPI_Irecv(&import_table->signal, 1, MPI_CHAR, target, 0, import_table->comm, import_table->request + i);
		}
#else
		MPI_Irecv(import_table->buffer + index, length, MPI_DOUBLE, target, 0, import_table->comm, import_table->request + i);
#endif
	}

#ifdef OUTPUT_LOG
	int byte;
	double start_time, end_time;
	byte = sizeof(double) * ( import_table->nnz + export_table->nnz );
	start_time = MPI_Wtime();
#endif

	//---------- send wait -------------
	for(i=0; i<export_table->neibpetot; i++){
		target = export_table->neibpe[i];

#ifndef USE_MPI_IPC
		if(export_table->neibpe_buffer[i] == 0 || NO_GPU_DIRECT_FLAG){
			MPI_Wait(export_table->request + i, export_table->status + i);
		}else{
			MPI_Wait(export_table->request + i, export_table->status + i);
			MPI_Irecv(&export_table->signal, 1, MPI_CHAR, target, 0, export_table->comm, export_table->request + i);
		}
#else
		MPI_Wait(export_table->request + i, export_table->status + i);
#endif
	}

	//---------- recv wait -------------
	for(i=0; i<import_table->neibpetot; i++){
		index  = import_table->ptr[ i ];
		length = import_table->ptr[i+1] - import_table->ptr[i];
		target = import_table->neibpe[i];

#ifndef USE_MPI_IPC
		if(import_table->neibpe_buffer[i] == 0 || NO_GPU_DIRECT_FLAG){
			MPI_Wait(import_table->request + i, import_table->status + i);
			my_cudaMemcpy(import_table->buffer + index, import_table->host_buffer + index, sizeof(double) * length, cudaMemcpyHostToDevice);
		}else{
			MPI_Wait(import_table->request + i, import_table->status + i);
			cudaMemcpy(import_table->buffer + index, import_table->neibpe_buffer[i], sizeof(double) * length, cudaMemcpyDeviceToDevice);
			MPI_Isend(&import_table->signal, 1, MPI_CHAR, target, 0, import_table->comm, import_table->request + i);
		}
#else
		MPI_Wait(import_table->request + i, import_table->status + i);
#endif
	}

	switch(opp){
	case COMM_MARGIN_EQL:
		put_into_the_data(data, import_table->buffer, import_table->dev_index, import_table->nnz);
		break;
	case COMM_MARGIN_ADD:
		add_into_the_data(data, import_table->buffer, import_table->dev_index, import_table->nnz);
		break;
	}

#ifndef USE_MPI_IPC
	//---------- send wait -------------
	for(i=0; i<export_table->neibpetot; i++){
		if(export_table->neibpe_buffer[i] == 0 || NO_GPU_DIRECT_FLAG){
		}else{
			MPI_Wait(export_table->request + i, export_table->status + i);
		}
	}

	//---------- recv wait -------------
	for(i=0; i<import_table->neibpetot; i++){
		if(import_table->neibpe_buffer[i] == 0 || NO_GPU_DIRECT_FLAG){
		}else{
			MPI_Wait(import_table->request + i, import_table->status + i);
		}
	}
#endif

#ifdef OUTPUT_LOG
	end_time  = MPI_Wtime();
	comm_log_input(byte, start_time, end_time);
#endif
}

template void comm_margin<COMM_MARGIN_EQL>(double data[], boader_zone *import_table, boader_zone *export_table);
template void comm_margin<COMM_MARGIN_ADD>(double data[], boader_zone *import_table, boader_zone *export_table);


void icomm_margin(double data[], boader_zone *import_table, boader_zone *export_table, comm_margin_opp opp)
{
	int i;
	int index, length, target;

	//---------- send margin -------------
	put_into_the_buffer(export_table->buffer, data, export_table->dev_index, export_table->nnz);

	for(i=0; i<export_table->neibpetot; i++){
		index  = export_table->ptr[ i ];
		length = export_table->ptr[i+1] - export_table->ptr[i];
		target = export_table->neibpe[i];

		if(export_table->neibpe_buffer[i] == 0 || NO_GPU_DIRECT_FLAG){
			my_cudaMemcpy(export_table->host_buffer + index, export_table->buffer + index, sizeof(double) * length, cudaMemcpyDeviceToHost);
			MPI_Isend(export_table->host_buffer + index, length, MPI_DOUBLE, target, 0, export_table->comm, export_table->request + i);
		}else{
			cudaThreadSynchronize();
			MPI_Isend(&export_table->signal, 1, MPI_CHAR, target, 0, export_table->comm, export_table->request + i);
		}
	}

	//---------- recv margin -------------
	for(i=0; i<import_table->neibpetot; i++){
		index  = import_table->ptr[ i ];
		length = import_table->ptr[i+1] - import_table->ptr[i];
		target = import_table->neibpe[i];
		
		if(import_table->neibpe_buffer[i] == 0 || NO_GPU_DIRECT_FLAG){
			MPI_Irecv(import_table->host_buffer + index, length, MPI_DOUBLE, target, 0, import_table->comm, import_table->request + i);
		}else{
			MPI_Irecv(&import_table->signal, 1, MPI_CHAR, target, 0, import_table->comm, import_table->request + i);
			MPI_Wait(import_table->request + i, import_table->status + i);
			cudaMemcpyAsync(import_table->buffer + index, import_table->neibpe_buffer[i], sizeof(double) * length, cudaMemcpyDeviceToDevice, import_table->stream);
		}
	}
}


void icomm_margin_wait(double data[], boader_zone *import_table, boader_zone *export_table, comm_margin_opp opp)
{
	int i;
	int index, length, target;

#ifdef OUTPUT_LOG
	int byte;
	double start_time, end_time;
	byte = sizeof(double) * ( import_table->nnz + export_table->nnz );
	start_time = MPI_Wtime();
#endif

	//---------- send wait -------------
	for(i=0; i<export_table->neibpetot; i++){
		target = export_table->neibpe[i];
		if(export_table->neibpe_buffer[i] == 0 || NO_GPU_DIRECT_FLAG){
			MPI_Wait(export_table->request + i, export_table->status + i);
		}else{
			MPI_Wait(export_table->request + i, export_table->status + i);
			MPI_Irecv(&export_table->signal, 1, MPI_CHAR, target, 0, export_table->comm, export_table->request + i);
		}
	}

	//---------- recv wait -------------
	for(i=0; i<import_table->neibpetot; i++){

		index  = import_table->ptr[ i ];
		length = import_table->ptr[i+1] - import_table->ptr[i];
		target = import_table->neibpe[i];

		if(import_table->neibpe_buffer[i] == 0 || NO_GPU_DIRECT_FLAG){
			MPI_Wait(import_table->request + i, import_table->status + i);
			my_cudaMemcpy(import_table->buffer + index, import_table->host_buffer + index, sizeof(double) * length, cudaMemcpyHostToDevice);
		}else{
			cudaStreamSynchronize(import_table->stream);
			MPI_Isend(&import_table->signal, 1, MPI_CHAR, target, 0, import_table->comm, import_table->request + i);
		}
	}


	switch(opp){
	case COMM_MARGIN_EQL:
		put_into_the_data(data, import_table->buffer, import_table->dev_index, import_table->nnz);
		break;
	case COMM_MARGIN_ADD:
		add_into_the_data(data, import_table->buffer, import_table->dev_index, import_table->nnz);
		break;
	}

	//---------- send wait -------------
	for(i=0; i<export_table->neibpetot; i++){
		if(export_table->neibpe_buffer[i] == 0 || NO_GPU_DIRECT_FLAG){
		}else{
			MPI_Wait(export_table->request + i, export_table->status + i);
		}
	}

	//---------- recv wait -------------
	for(i=0; i<import_table->neibpetot; i++){
		if(import_table->neibpe_buffer[i] == 0 || NO_GPU_DIRECT_FLAG){
		}else{
			MPI_Wait(import_table->request + i, import_table->status + i);
		}
	}

#ifdef OUTPUT_LOG
	end_time  = MPI_Wtime();
	comm_log_input(byte, start_time, end_time);
#endif
}



bool my_cuda_vector_less_with_mpi(double *dx, double *dy, int n, MPI_Comm solver_comm)
{
	bool flag;
//	char local_mpi_bool, mpi_bool;
	int local_mpi_bool, mpi_bool;
	flag = my_cuda_vector_less(dx, dy, n);
	local_mpi_bool = flag ? 1 : 0;
//	MPI_Allreduce(&local_mpi_bool, &mpi_bool, 1, MPI_CHAR, MPI_LAND, solver_comm);
	MPI_Allreduce(&local_mpi_bool, &mpi_bool, 1, MPI_INT, MPI_LAND, solver_comm);
	flag = mpi_bool ? true : false;

	return flag;
}


double my_cuda_dot_with_mpi(double *a, double *b, int n, MPI_Comm solver_comm)
{
	double local_dot, dot;
	local_dot = my_cuda_dot(a, b, n);
	MPI_Allreduce(&local_dot, &dot, 1, MPI_DOUBLE, MPI_SUM, solver_comm);
	return dot;
}


double get_two_norm_residual_with_mpi(lu_crs &a, double *x, double *b, MPI_Comm solver_comm)
{
	double local_residual, residual;

	local_residual = get_two_norm_residual(a, x, b);
	local_residual = local_residual * local_residual;
	MPI_Allreduce(&local_residual, &residual, 1, MPI_DOUBLE, MPI_SUM, solver_comm);
	residual = sqrt(residual);

	return residual;
}


double get_two_norm_residual_with_mpi(lu_icrs &a, double *x, double *b, MPI_Comm solver_comm)
{
	double local_residual, residual;

	local_residual = get_two_norm_residual(a, x, b);
	local_residual = local_residual * local_residual;
	MPI_Allreduce(&local_residual, &residual, 1, MPI_DOUBLE, MPI_SUM, solver_comm);
	residual = sqrt(residual);
	return residual;
}

