#include "my_cuda.h"
#include "smoother.h"

#include <cstring>
#include <iostream>
#include <map>

using std::cout;
using std::endl;

using std::map;

/*
 * must set b_n, b_cnt, b_ind, order
 */

typedef struct
{
	double color_threshold;
}mcgs_tmp;

map<v_node*, mcgs_tmp> mcgs_map;

void multi_color_gauss_seidel_setup(v_node &v)
{
	double color_threshold;

	if(v.level != 0){
		color_threshold = 0.05;
	}else{
		color_threshold = 0.00;
	}

	multi_color_gauss_seidel_setup(v.a, v.da, color_threshold, &v.b_n, &v.b_cnt, &v.b_ind, &v.order);
}


void multi_color_gauss_seidel(v_node &v, double omega)
{
	multi_color_gauss_seidel(v.da, v.dx, v.db, omega);
//	multi_color_gauss_seidel(v.da, v.dx, v.db, omega, &v.area_outer, &v.area_inner);
}


typedef struct
{
	double *x_buffer;
}jacobi_tmp;

map<v_node*, jacobi_tmp> jacobi_map;

void jacobi_setup(v_node &v)
{
	int i, dnp;
	jacobi_tmp tmp;

	jacobi_setup(v.a, v.da, &v.b_n, &v.b_cnt, &v.b_ind, &v.order);

	dnp = v.a.np - v.a.n;
	for(i=0; i<v.da.size(); i++){
		dnp += v.da[i].np;
	}

	my_cudaMalloc((void**)&tmp.x_buffer, sizeof(double) * dnp, "jacobi x_buffer");
	jacobi_map[ &v ] = tmp;
}


void jacobi(v_node &v, double omega)
{
	double *old_x, *new_x;

	old_x = v.dx;
	new_x = jacobi_map[&v].x_buffer;

	jacobi(v.da, old_x, new_x, v.db, omega);
	my_cudaMemcpy(old_x, new_x, sizeof(double) *v.dnp, cudaMemcpyDeviceToDevice);
}


void jacobi_finalize(v_node &v)
{
	my_cudaFree(jacobi_map[&v].x_buffer);
}


typedef struct
{
	double gamma;
	double *t;
	double *dat;
	double *r;
	double *p;
	double *s;
	double *dr;
	double *dx;
	double *buffer;
}idr_tmp;

map<v_node*, idr_tmp> idr_map;

void idr_jacobi_setup(v_node &v)
{
	int i, dnp;
	idr_tmp tmp;

	idr_jacobi_setup(v.a, v.da, &v.b_n, &v.b_cnt, &v.b_ind, &v.order);

	dnp = v.a.np - v.a.n;
	for(i=0; i<v.da.size(); i++){
		dnp += v.da[i].np;
	}

	my_cudaMalloc((void**)&tmp.buffer, sizeof(double) * dnp * 7, "idr_jacobi temp_buffer");
	tmp.t  = tmp.buffer + 0 * dnp;
	tmp.dat= tmp.buffer + 1 * dnp;
	tmp.r  = tmp.buffer + 2 * dnp;
	tmp.p  = tmp.buffer + 3 * dnp;
	tmp.s  = tmp.buffer + 4 * dnp;
	tmp.dr = tmp.buffer + 5 * dnp;
	tmp.dx = tmp.buffer + 6 * dnp;
	my_cudaMemset(tmp.buffer, 0, sizeof(double) * dnp * 7);

	idr_map[ &v ] = tmp;
}

void idr_jacobi_init(v_node &v)
{
	int n = v.da[0].n;
	double *p = idr_map[&v].p;
	idr_jacobi_init(v.da, v.dx, v.db, idr_map[&v].r, idr_map[&v].p, &idr_map[&v].gamma);
}

void idr_jacobi(v_node &v, double omega)
{
	idr_jacobi(v.da, v.dx, v.db, idr_map[&v].t, idr_map[&v].dat, idr_map[&v].r, idr_map[&v].p, idr_map[&v].s, idr_map[&v].dr, idr_map[&v].dx, &idr_map[&v].gamma, omega, v.comm);
}


void idr_jacobi_finalize(v_node &v)
{
	my_cudaFree( idr_map[&v].buffer );

	idr_map.erase( &v );
}


typedef struct
{
	double *lu;
	int *p;
}lu_tmp;

map<v_node*, lu_tmp> lu_map;

void lu_decomposition_setup(v_node &v)
{
	int n, np;
	lu_tmp tmp;

	n  = v.a.n;
	np = v.a.np;

	tmp.lu = new double[np * np];
	tmp.p  = new int[np];
	
	lu_decomposition_setup(&v.a, tmp.lu, tmp.p);

	lu_map[ &v ] = tmp;
	
	v.b_n   = 1;
	v.b_cnt = new int[1];
	v.b_ind = new int[n];

	v.b_cnt[0] = n;
	memset(v.b_ind, 0, sizeof(int) * n);

	convert_lu_crs_to_lu_icrs(v.a, v.da, v.b_n, v.b_cnt, v.b_ind, &v.order);
}


void lu_decomposition(v_node &v, double omega)
{
	int n, np, dnp;
	int *order;
	double *b, *x, *db, *dx;

	n   = v.n;
	np  = v.np;
	dnp = v.dnp;
	db  = v.db;
	dx  = v.dx;
	
	order = v.order;

	x = v.x;
	b = v.b;

	my_cuda_device_to_host_array(b, db, np, dnp, order);
	
	lu_decomposition(lu_map[&v].lu, x, b, lu_map[&v].p, np);
	
	my_cuda_host_to_device_array(dx, x, dnp, np, order);

}


void lu_decomposition_finalize(v_node &v)
{
	delete[] lu_map[&v].lu;
	delete[] lu_map[&v].p;

	lu_map.erase( &v );
}

