!C   Copyright (C) 2010 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY THE SCALABLE SOFTWARE INFRASTRUCTURE PROJECT
!C   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
!C   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
!C   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE SCALABLE SOFTWARE INFRASTRUCTURE
!C   PROJECT BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
!C   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
!C   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
!C   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
!C   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
!C   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
!C   POSSIBILITY OF SUCH DAMAGE.
module terminal_condition
  include 'precision.inc'
  integer(kind=kint)   :: terminal_kind = 0, conv_vec_size = 0
  REAL(kind=kreal), pointer :: conv_criterion_vec(:)
  REAL(kind=kreal)          :: conv_criterion_val=1.0E-8
  !C BNRM2 is the sum of squares of rhs vector in the case of terminal_kind == 0
  !C BNRM2 is the inf-norm of rhs vector in the case of terminal_kind == 1
  REAL(kind=kreal)          :: BNRM2 
  LOGICAL :: conv_vec_REORDERED_FL = .false.
  
contains
  subroutine BNRM2_set(B, SOLVER_COMM, N)
    !$ USE omp_lib
    implicit none
    include 'mpif.h'
    
    REAL(kind=kreal)   :: B(:), BNRM20, WVAL
    INTEGER(kind=kint) :: SOLVER_COMM, ierr, i, N

    if(terminal_kind == 0) then
       BNRM20 = 0.d0
       !$omp parallel do reduction(+:BNRM20)
       do i= 1, N
          BNRM20 = BNRM20 + B(i)**2
       enddo
       !$omp end parallel do
       
       call MPI_allREDUCE (BNRM20, BNRM2, 1, MPI_DOUBLE_PRECISION,       &
            &                    MPI_SUM, SOLVER_COMM, ierr)
    else if(terminal_kind == 1) then
       BNRM20 = 0.d0
       !$omp parallel do reduction(max:BNRM20) private(WVAL)
       do i= 1, N
          WVAL = dabs(B(i))
          if(BNRM20 < WVAL) BNRM20 = WVAL
       enddo
       !$omp end parallel do
       call MPI_allREDUCE (BNRM20, BNRM2, 1, MPI_DOUBLE_PRECISION,       &
            &                    MPI_MAX, SOLVER_COMM, ierr)
    else 
       stop 'unknown terminal_kind'
    end if

    if(BNRM2 == 0) BNRM2 = 1.0
    
  end subroutine BNRM2_set





  SUBROUTINE terminal_cond_vec_set(res_vec, N)
    !$ USE omp_lib
    implicit none
    
    integer(kind=kint), intent(in) :: N
    real(kind=kreal), intent(in)   :: res_vec(:)
    logical, save :: init_flag=.true.
    integer(kind=kint) :: i


    if(init_flag) then
       nullify(conv_criterion_vec)
       init_flag  = .false.
    end if
    terminal_kind=1
    if(ASSOCIATED(conv_criterion_vec)) then
#ifdef DEBUG
       write(*,*) "Convergence Criterion vector is already allocated"
#endif
       if(conv_vec_size < N) then
          deallocate(conv_criterion_vec)
          allocate(conv_criterion_vec(N))
       end if
    else 
       allocate(conv_criterion_vec(N))
    end if

    conv_vec_size = N
    !$omp parallel do 
    do i = 1, N
       conv_criterion_vec(i) = res_vec(i)
    end do
    !$omp end parallel do

    
    conv_vec_REORDERED_FL = .false.
  END SUBROUTINE terminal_cond_vec_set

  SUBROUTINE terminal_cond_val_set(res_val)
    implicit none
    real(kind=kreal), intent(in) ::  res_val

    terminal_kind = 0
    conv_criterion_val = res_val

  END SUBROUTINE terminal_cond_val_set

  SUBROUTINE terminal_cond_check(CONVERGENCE_FL, resid, residual_vec, N, SOLVER_COMM)
    use problem_matrix,   only: REORDERED_FL,nds_index_aft_dst,order_for_dst

    !$ use omp_lib
    
    implicit none
    include 'mpif.h'
    logical, intent(out) :: CONVERGENCE_FL
    REAL(kind=kreal), intent(out) :: RESID
    REAL(kind=kreal), intent(in) :: residual_vec(:)
    integer(kind=kint), intent(in) :: SOLVER_COMM, N

    real(kind=kreal)   :: RNRM, RNRM2, infnrm_data0(2), infnrm_data(2), v, w 
    real(kind=kreal),pointer   :: conv_criterion_vec_t(:)
    integer(kind=kint) :: i, ierr
    logical :: conv_fl

    if(terminal_kind==1) then
       call reorder_conv_vec(N)
!!$       !C reordering of conv_criterion_vec
!!$       if(REORDERED_FL .and. .not.conv_vec_REORDERED_FL) then
!!$          
!!$          !C distribute conv_criterion_vec vector.
!!$          !C size of the vector is changed from conv_vec_size -> N
!!$          allocate(conv_criterion_vec_t(N))
!!$          call distribute_vector(nds_index_aft_dst, order_for_dst, SOLVER_COMM, &
!!$               & conv_criterion_vec, conv_vec_size, conv_criterion_vec_t)    
!!$          deallocate(conv_criterion_vec)
!!$          conv_criterion_vec => conv_criterion_vec_t
!!$
!!$          !C reset conv_vec_size
!!$          conv_vec_size = N
!!$          conv_vec_REORDERED_FL = .true.
!!$       end if

       conv_fl = .true.
       v = 0
       !$omp parallel do reduction(.and.: conv_fl) reduction(max: v) private(w)
       do i = 1, N
          w = dabs(residual_vec(i))
          conv_fl = conv_fl .and. (w <= conv_criterion_vec(i))
          if(v < w) v = w
       end do
       !$omp end parallel do
       
       call MPI_ALLREDUCE(conv_fl, CONVERGENCE_FL, 1, MPI_LOGICAL, MPI_LAND, &
            & SOLVER_COMM, ierr)
#ifdef PRINT_REZ
       call MPI_ALLREDUCE(v, RESID, 1, MPI_DOUBLE_PRECISION, MPI_MAX, & 
            & SOLVER_COMM, ierr)
       RESID = RESID / BNRM2
#endif
       
       
    else if(terminal_kind==0) then
       RNRM = 0.0

       !$omp parallel do  reduction(+: RNRM)
       do i = 1, N
          RNRM = RNRM+residual_vec(i) ** 2
       end do

       !$omp end parallel do 
       call MPI_ALLREDUCE(RNRM, RNRM2, 1, MPI_DOUBLE_PRECISION, MPI_SUM, &
            & SOLVER_COMM, ierr)
       RESID = dsqrt(RNRM2/BNRM2)
       CONVERGENCE_FL = (RESID <= conv_criterion_val)
       
    end if
    
    

  END SUBROUTINE terminal_cond_check
  
  subroutine reorder_conv_vec(N)
    use problem_matrix,   only: REORDERED_FL,nds_index_aft_dst,order_for_dst, SOLVER_COMM,NB
    use matrix_dist_mtom, only: distribute_vector
    implicit none
    integer(kind=kint) :: N
    real(kind=kreal),pointer   :: conv_criterion_vec_t(:)
    
    !C reordering of conv_criterion_vec
    if(REORDERED_FL .and. .not.conv_vec_REORDERED_FL) then
       
       !C distribute conv_criterion_vec vector.
       !C size of the vector is changed from conv_vec_size -> N
       allocate(conv_criterion_vec_t(N))
       call distribute_vector(nds_index_aft_dst, order_for_dst, SOLVER_COMM, &
            & conv_criterion_vec, conv_vec_size, conv_criterion_vec_t,NB)    
       deallocate(conv_criterion_vec)
       conv_criterion_vec => conv_criterion_vec_t
       
       !C reset conv_vec_size
       conv_vec_size = N
       conv_vec_REORDERED_FL = .true.
    end if
    
  end subroutine reorder_conv_vec

  subroutine terminal_cond_vec_clear()
    implicit none
    if(associated(conv_criterion_vec)) deallocate(conv_criterion_vec)
    conv_vec_size = 0
    conv_vec_REORDERED_FL = .false.    
  end subroutine terminal_cond_vec_clear

end module terminal_condition
